# coding: utf-8
import socket, json as _json


class HttpTestResponse:

    def __init__(self, message_list):
        self.content = b"".join(message_list)
        self.http_version, self.status_code, self.status_code_msg = None, None, None
        self.headers = {}
        self.body = None
        message_list = [msg.decode() for msg in message_list]
        index = 0
        for msg in message_list:
            index += 1
            if index == 1:
                head = msg.strip()
                self.http_version, self.status_code, self.status_code_msg = head.split(" ", 2)
                self.status_code = int(self.status_code)
            elif msg == "\r\n":
                self.body = "\r\n".join(message_list[index:])
                break
            else:
                k, v = msg.strip().split(":", 1)
                self.headers[k] = v.strip()

    def __str__(self):
        headers = []
        for k, v in self.headers.items():
            headers.append(f"{k}: {v}")
        header = "\r\n".join(headers)
        return f"{self.http_version} {self.status_code} {self.status_code_msg}\r\n{header}\r\n\r\n{self.body}"


class HttpTestRequest:

    def __init__(self, app):
        self.app = app
        self.method = None
        self.path = None
        self.body = None
        self.headers = dict()
        self.app.build()

    def request(self, method, route, headers=None, json=None, data=""):
        self.method = method
        self.path = route
        self.set_header("Pywss-Http-Test", "0.0.1")
        self.set_headers(headers or {})
        if json:
            self.set_headers({"Content-Type": "application/json"})
            data = _json.dumps(json, ensure_ascii=False)
        if isinstance(data, str):
            self.body = data
        elif isinstance(data, dict):
            self.set_headers({"Content-Type": "application/x-www-form-urlencoded"})
            self.body = "&".join([f"{k}={v}" for k, v in data.items()])
        return self.build()

    def set_header(self, k, v):
        header = []
        for key in k.split("-"):
            header.append(key[0].upper() + key[1:].lower())
        self.headers["-".join(header)] = v
        return self

    def set_headers(self, headers):
        for k, v in headers.items():
            self.set_header(k, v)
        return self

    def build(self) -> HttpTestResponse:
        if self.body:
            self.headers.setdefault("Content-Length", len(self.body.encode()))
        header_list = []
        for k, v in self.headers.items():
            header_list.append(f"{k}: {v}")
        header = "\r\n".join(header_list)
        req_message = f"{self.method} {self.path} HTTP/1.1\r\n{header}\r\n\r\n{self.body}"

        s, c = socket.socketpair()
        with s, c:
            c.sendall(req_message.encode())
            c.sendall("GET / HTTP/1.1\r\nConnection: close\r\n\r\n".encode())
            self.app.handler_request(s, None)
            resp = c.makefile("rb", -1)
            return HttpTestResponse(resp.readlines())

    def get(self, route, headers=None, json=None, data=""):
        return self.request("GET", route, headers, json, data)

    def post(self, route, headers=None, json=None, data=""):
        return self.request("POST", route, headers, json, data)

    def head(self, route, headers=None, json=None, data=""):
        return self.request("HEAD", route, headers, json, data)

    def put(self, route, headers=None, json=None, data=""):
        return self.request("PUT", route, headers, json, data)

    def delete(self, route, headers=None, json=None, data=""):
        return self.request("DELETE", route, headers, json, data)

    def patch(self, route, headers=None, json=None, data=""):
        return self.request("PATCH", route, headers, json, data)

    def options(self, route, headers=None, json=None, data=""):
        return self.request("OPTIONS", route, headers, json, data)



# coding: utf-8
import pywss

from pywss.constant import *


def NewCORSHandler(
        allow_origins: tuple = ("*",),
        allow_methods: tuple = ("*",),
        allow_headers: tuple = ("*",),
        allow_credentials: bool = True,
):
    allowOrigins = ",".join(allow_origins)
    allowMethods = ",".join(allow_methods)
    allowHeaders = ",".join(allow_headers)
    allowCredentials = "true" if allow_credentials else "false"

    def corsHandler(ctx: pywss.Context):
        ctx.set_header(HeaderAccessControlAllowOrigin, allowOrigins)
        ctx.set_header(HeaderAccessControlAllowMethods, allowMethods)
        ctx.set_header(HeaderAccessControlAllowHeaders, allowHeaders)
        ctx.set_header(HeaderAccessControlAllowCredentials, allowCredentials)
        if ctx.method == MethodOptions:
            return
        ctx.next()

    return corsHandler



# coding: utf-8
import os
import pywss

from pywss.constant import StatusNotFound, StatusServiceUnavailable, MethodHead


def NewStaticHandler(root, default="application/octet-stream", limit=100):
    textHtml = tuple("html,htm,shtml".split(","))
    textPlain = tuple("py,go,md,txt,log".split(","))
    textCss = tuple("css".split(","))
    textXml = tuple("xml".split(","))
    applicationXJavascript = tuple("js".split(","))
    applicationJson = tuple("json,yml,yaml".split(","))
    imagePng = tuple("jpg,jpeg,png,gif,avif,svg,svgz,wbmp,webp,ico,bmp".split(","))

    def staticHandler(ctx: pywss.Context):
        if ctx.headers.get("Content-Range", None):
            ctx.set_status_code(StatusServiceUnavailable)
            ctx.write("Not Support Header Content-Range")
            return
        _path = f"{ctx.method}/{ctx.route.strip('/')}"
        file = os.path.join(root, *_path.replace(ctx._route, "").strip("/").split("/"))
        if not os.path.exists(file):
            ctx.set_status_code(StatusNotFound)
            return
        if os.path.isdir(file):
            if ctx.route[-1] != "/":
                ctx.redirect(ctx.route + "/")
                return
            ctx.set_content_type("text/html; charset=UTF-8")
            ctx.write(html_template(file, limit))
            return
        if file.endswith(textHtml):
            ctx.set_content_type("text/html; charset=UTF-8")
        elif file.endswith(textPlain):
            ctx.set_content_type("text/plain; charset=UTF-8")
        elif file.endswith(textCss):
            ctx.set_content_type("text/css")
        elif file.endswith(applicationXJavascript):
            ctx.set_content_type("application/javascript")
        elif file.endswith(applicationJson):
            ctx.set_content_type("application/json")
        elif file.endswith(textXml):
            ctx.set_content_type("text/xml")
        elif file.endswith(imagePng):
            ctx.set_content_type("image/png")
        else:
            ctx.set_content_type(default)
        if ctx.method == MethodHead:
            with open(file, "rb") as f:
                ctx.set_content_length(os.stat(f.fileno())[6])
            return
        ctx.write_file(open(file, "rb"))

    return staticHandler


def html_template(path, limit):
    count = 0
    dirs = []
    files = []
    for el in os.scandir(path):
        name = el.name
        if len(name) > 8:
            name = name[:8] + "..."
        if el.is_dir():
            dirs.append(f'<a class="box box-dir" title="目录：{el.name}" href="{el.name}/">{name}</a>')
        else:
            files.append(f'<a class="box box-file" '
                         f'title="文件名：{el.name}\n文件大小：{el.stat().st_size}" href="{el.name}">{name}</a>')
        count += 1
        if count >= limit:
            break
    data = "\n".join(dirs) + "\n".join(files)
    return f"""<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Pywss Server</title>
    <style>
        * {{
            margin: 0;
            padding: 0;
        }}
        .container {{
            width: 100vw;
            overflow-x: hidden;
            display: flex;
            flex-wrap: wrap;
        }}
        .box {{
            margin: 15px;
            width: 100px;
            height: 100px;
            border-radius: 30px;
            display: flex;
            align-items: center;
            justify-content: center;
            overflow: hidden;
            font-weight: bold;
            color: #f1f1f1;
            text-decoration: none;
        }}
        .box:hover {{
            box-shadow: 0 0 8px 6px #7f7a7a;
            cursor: pointer;
        }}
        .box-dir {{
            background: #4495dc;
        }}
        .box-file {{
            background: #656668;
        }}
    </style>
</head>
<body>
<div class="container">
    {data}
</div>
</body>
</html>
"""



# coding: utf-8
import json
import struct
import base64
import hashlib
import warnings

MAGIC_STRING = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'
RESPONSE_TEMPLATE = "HTTP/1.1 101 Switching Protocols\r\n" \
                    "Upgrade: websocket\r\n" \
                    "Connection: Upgrade\r\n" \
                    "Sec-WebSocket-Accept: %s\r\n\r\n"


def WebSocketUpgrade(ctx):
    if ctx.headers.get("Upgrade") != "websocket":
        return "invalid websocket request"
    secKey = ctx.headers.get("Sec-WebSocket-Key")
    if not secKey:
        return "invalid websocket key"
    secResp = _createWebSocketResponse(secKey)
    ctx.fd.sendall(secResp)

    def ws_write(body):
        if isinstance(body, bytes):
            ctx.fd.sendall(_websocketEncodeMsg(body))
        elif isinstance(body, str):
            ctx.fd.sendall(_websocketEncodeMsg(body.encode("utf-8")))
        elif isinstance(body, (dict, list)):
            ctx.fd.sendall(_websocketEncodeMsg(json.dumps(body, ensure_ascii=False).encode("utf-8")))

    ctx.ws_read = lambda: _websocketRead(ctx.rfd)
    ctx.ws_write = ws_write
    return None


def _createWebSocketResponse(secKey):
    secKey = secKey + MAGIC_STRING
    secKey = hashlib.sha1((secKey).encode('utf-8')).digest()
    secKey = base64.b64encode(secKey).decode('utf-8')
    return bytes(RESPONSE_TEMPLATE % secKey, encoding='utf-8')


def _websocketRead(sock) -> bytes:
    response = sock.read(2)
    if len(response) != 2:
        return b""
    length = response[1] & 0b1111111
    if length is 0b1111110:
        response += sock.read(2)
        _, data_length = struct.unpack('!BH', response[1:4])
    elif length is 0b1111111:
        response += sock.read(8)
        _, data_length = struct.unpack('!BQ', response[1:10])
    else:
        data_length = length
    data_length += 4
    loop = data_length // 4096
    remain = data_length % 4096
    while loop:
        response += sock.read(4096)
        loop -= 1
    response += sock.read(remain)
    return _websocketDecodeMsg(response)


def _websocketDecodeMsg(data) -> bytes:
    payload_len = data[1] & 0b1111111
    if payload_len is 0b1111110:
        mask = data[4:8]
        decoded = data[8:]
    elif payload_len is 0b1111111:
        mask = data[10:14]
        decoded = data[14:]
    else:
        mask = data[2:6]
        decoded = data[6:]
    return bytes(bytearray([decoded[i] ^ mask[i % 4] for i in range(len(decoded))]))


def _websocketEncodeMsg(msg_bytes, token=b"\x81") -> bytes:
    length = len(msg_bytes)
    if length <= 125:
        token += struct.pack("B", length)
    elif length <= 65535:
        token += struct.pack("!BH", 126, length)
    else:
        token += struct.pack("!BQ", 127, length)
    return token + msg_bytes
